package edu.kit.ipd.descartes.ql.core.engine.aggregates;

import static edu.kit.ipd.descartes.ql.core.engine.util.AggregateHelper.mergeRequiredMetrics;
import static edu.kit.ipd.descartes.ql.core.engine.util.EntityMappingHelper.buildEntityIndex;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.eclipse.emf.common.util.EList;

import edu.kit.ipd.descartes.ql.connector.util.DQLLogger;
import edu.kit.ipd.descartes.ql.core.engine.util.EntityMappingHelper;
import edu.kit.ipd.descartes.ql.models.mapping.domain.DomainFactory;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Entity;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Probe;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.EntityMapping;

/**
 * Implementation of AggregateController
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public class AggregateControllerImpl implements AggregateController {
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private final List<AggregateProcessor> processors = new ArrayList<AggregateProcessor>();
	private final Map<String, Set<String>> requiredMetrics = new HashMap<String, Set<String>>();
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public AggregateControllerImpl() {
		super();
	}

	public boolean hasAggregates() {
		return !processors.isEmpty();
	}

	public void postProcess(final EntityMapping response) {
		if (log.isDebugEnabled()) {
			log.debug("Starting post-processing of aggregates");
		}

		// TODO: Comment needs to be changed. Indexes are now pre-built.
		// At this stage all probes required as part of aggregates can be
		// removed. The removal process is based on an index for entities and
		// for each entity an index for its probes. Although the indexing comes
		// at the cost of an overhead, it is more reasonable than double-looped
		// comparisons of entity identifiers and metric names. If the amount of
		// requested entities and metrics increases, using indexing will be an
		// advantage. To reason the overhead, the data structures for building
		// the indexes are initialized with their final capacity. No additional
		// memory management is needed.
		final Map<String, Entity> entityIndex = buildEntityIndex(response);
		final Map<String, Map<String, Probe>> probeIndex = EntityMappingHelper.buildProbeIndex(response);

		for (final AggregateProcessor processor : processors) {
			processor.postProcess(response, entityIndex, probeIndex);
		}

		if (log.isDebugEnabled()) {
			log.debug("Starting to remove previously added probes for"
					+ " calculating aggregates that are no longer used");
		}

		// Remove all probes (at this time DecimalResults actually) that have
		// been inserted into the original request for calculating aggregates.
		// The requiredMetrics index contains the merged set of required
		// metrics.
		for (final String identifier : requiredMetrics.keySet()) {
			final Entity entity = entityIndex.get(identifier);
			final EList<Probe> probes = entity.getProbes();

			removeProbes(identifier, probes, probeIndex.get(identifier));
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateController#
	 * preProcess(edu.kit.ipd.descartes.ql.models.mapping.mapping.EntityMapping,
	 * java.util.Map)
	 */
	public void preProcess(final EntityMapping request, final Map<String, Entity> aliasMap) {
		if (log.isDebugEnabled()) {
			log.debug("Starting pre-processing of aggregates");
		}

		for (final AggregateProcessor processor : processors) {
			mergeRequiredMetrics(requiredMetrics, processor.preProcess(request, aliasMap));
		}

		final Map<String, Entity> entityIndex = buildEntityIndex(request);

		for (final String identifier : requiredMetrics.keySet()) {
			final Entity entity = entityIndex.get(identifier);
			final EList<Probe> probes = entity.getProbes();

			for (final String metricName : requiredMetrics.get(identifier)) {
				final Probe probe = domainFactory.createProbe();
				probe.setMetricName(metricName);
				probes.add(probe);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateController#
	 * registerAggregate
	 * (edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateProcessor)
	 */
	public void registerAggregate(final AggregateProcessor processor) {
		if (log.isDebugEnabled()) {
			log.debug("Registering a new aggregate processor: " + processor);
		}

		if (processor == null) {
			return;
		}

		processors.add(processor);
	}

	private void removeProbes(final String identifier, final EList<Probe> probes,
			final Map<String, Probe> probesOfEntity) {
		for (final String metricName : requiredMetrics.get(identifier)) {
			final Probe probe = probesOfEntity.get(metricName);
			probes.remove(probe);
		}
	}
}
