package edu.kit.ipd.descartes.ql.core.engine.interpreter;

import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceReference;

import edu.kit.ipd.descartes.ql.connector.ConnectorProvider;
import edu.kit.ipd.descartes.ql.connector.ModelStructureQueryConnector;
import edu.kit.ipd.descartes.ql.connector.PerformanceIssueQueryConnector;
import edu.kit.ipd.descartes.ql.connector.PerformanceMetricsQueryConnector;
import edu.kit.ipd.descartes.ql.connector.util.DQLLogger;
import edu.kit.ipd.descartes.ql.core.engine.Activator;
import edu.kit.ipd.descartes.ql.core.registry.ConnectorRegistry;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.EntityMapping;

public abstract class AbstractQueryInterpreter<Q, qC> implements QueryInterpreter {
	protected Logger log = DQLLogger.getLogger(this.getClass().getName());
	private BundleContext context = Activator.getContext();
	private qC queryConnector = null;
	private List<EntityMapping> results = null;
	private Q query = null;
	private String modelFamily = null;
	private ServiceReference<ConnectorRegistry> registryServiceRef;
	private final Class<qC> myQueryConnectorClass = getQueryConnectorClass();

	public AbstractQueryInterpreter(final Q query) {
		this.query = query;
	}

	public BundleContext getContext() {
		return context;
	}

	public String getModelFamily() {
		return modelFamily;
	}

	public Q getQuery() {
		return query;
	}

	public qC getQueryConnector() {
		return queryConnector;
	}

	protected abstract Class<qC> getQueryConnectorClass();

	// TODO: Should be implemented with type-safety
	@SuppressWarnings("unchecked")
	private qC getQueryConnectorInstance(final ConnectorProvider connectorProvider) {
		log.debug("Trying to create QueryConnector of type " + myQueryConnectorClass.getName());

		if (myQueryConnectorClass == ModelStructureQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createModelStructureQueryConnectorInstance();
		} else if (myQueryConnectorClass == PerformanceMetricsQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createPerformanceMetricsQueryConnectorInstance();
		} else if (myQueryConnectorClass == PerformanceIssueQueryConnector.class) {
			queryConnector = (qC) connectorProvider.createPerformanceIssueQueryConnectorInstance();
		}

		return queryConnector;
	}

	public EntityMapping getResult() {
		if (results == null) {
			return null;
		}

		return results.get(0);
	}

	public List<EntityMapping> getResults() {
		return results;
	}

	public boolean hasMultipleResults() {
		if (results == null || results.size() < 2) {
			return false;
		}

		return true;
	}

	public boolean initQueryConnector() {
		// No null check, as the registry is a dependency and must be available
		// if this bundle is active
		final ConnectorRegistry connectorRegistry = lookupConnectorRegistryInstance();

		final ServiceReference<ConnectorProvider> connectorProviderRef = lookupConnectorProviderReference(
				connectorRegistry, getModelFamily());

		// Might happen even caused by a typo in the query...
		if (connectorProviderRef == null) {
			log.error("No ConnectorProvider found for family " + getModelFamily() + ", aborting.");
			return false;
		}

		final ConnectorProvider connectorProvider = context.getService(connectorProviderRef);

		// Might happen due to dynamics in the run-time and no direct bundle
		// dependency
		if (connectorProvider == null) {
			log.error("Could not instantiate the ConnectorProvider, aborting.");
			return false;
		}

		queryConnector = getQueryConnectorInstance(connectorProvider);

		// Erroneous implementations of the ConnectorProvider might cause this
		if (queryConnector == null) {
			log.error("Although advertised, the ConnectorProvider can not offer a " + myQueryConnectorClass
					+ ", aborting.");
			return false;
		}

		return true;
	}

	protected ServiceReference<ConnectorProvider> lookupConnectorProviderReference(final ConnectorRegistry registry,
			final String family) {
		return registry.requestServiceReference(family, myQueryConnectorClass);
	}

	private ConnectorRegistry lookupConnectorRegistryInstance() {
		registryServiceRef = context.getServiceReference(ConnectorRegistry.class);
		return context.getService(registryServiceRef);
	}

	public void setContext(final BundleContext context) {
		this.context = context;
	}

	public void setModelFamily(final String modelFamily) {
		this.modelFamily = modelFamily;
	}

	public void setQuery(final Q query) {
		this.query = query;
	}

	public void setResult(final EntityMapping result) {
		if (results == null) {
			results = new ArrayList<EntityMapping>();
		}

		results.add(result);
	}

	public void setResults(final List<EntityMapping> results) {
		this.results = results;
	}
}
