package edu.kit.ipd.descartes.ql.core.engine.interpreter;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import edu.kit.ipd.descartes.ql.connector.ModelStructureQueryConnector;
import edu.kit.ipd.descartes.ql.lang.descartesQL.EntityReference;
import edu.kit.ipd.descartes.ql.lang.descartesQL.EntityReferenceClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.EntityType;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListDoFQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListEntitiesQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListMetricsQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListResourcesQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ListServicesQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelFamily;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelReferenceClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelStructureQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.UsingClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.WithClause;
import edu.kit.ipd.descartes.ql.models.mapping.domain.DomainFactory;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Entity;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Resource;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Service;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.EntityMapping;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.MappingFactory;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.impl.MappingFactoryImpl;

/**
 * Interpreter for LIST queries
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public class ModelStructureQueryInterpreter extends
		AbstractQueryInterpreter<ModelStructureQuery, ModelStructureQueryConnector> {

	private final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public ModelStructureQueryInterpreter(final ModelStructureQuery query) {
		super(query);
	}

	@Override
	protected Class<ModelStructureQueryConnector> getQueryConnectorClass() {
		return ModelStructureQueryConnector.class;
	}

	/**
	 * Interpret query as LIST DOF
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListDoFQuery(final ListQuery query) {
		final UsingClause usingClause = query.getUsingClause();
		final ListDoFQuery doFQuery = (ListDoFQuery) query.getQueryType();
		final ModelReferenceClause modelRefClause = usingClause.getModelReference();

		if (!getQueryConnector().supportsDoFRequests()) {
			log.error("Connector does not support DoF requests, aborting.");
			return false;
		}

		final MappingFactory mappingFactory = MappingFactoryImpl.init();
		final EntityMapping request = mappingFactory.createEntityMapping();

		request.setModelLocation(modelRefClause.getModel().getLocation());

		final WithClause withClause = doFQuery.getWithClause();

		if (withClause != null) {
			request.setDoFModelLocation(withClause.getModelReference().toString());
		}

		setResult(getQueryConnector().requestDoF(request));

		return true;
	}

	/**
	 * Interpret query as LIST ENTITIES
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListEntitiesQuery(final ListQuery query) {
		// TODO: Implement
		// FilterClause filterClause = ((ListQuery) query).getFilterClause();
		final UsingClause usingClause = query.getUsingClause();
		final ModelReferenceClause modelRefClause = usingClause.getModelReference();

		if (!getQueryConnector().supportsEntityRequests()) {
			log.error("Connector does not support entity requests, aborting.");
			return false;
		}

		final EntityMapping request = mappingFactory.createEntityMapping();
		request.setModelLocation(modelRefClause.getModel().getLocation());

		// TODO: Interpret FilterClause

		// Delegate query to connector
		setResult(getQueryConnector().requestEntities(request));

		return true;
	}

	/**
	 * Interpret query as LIST METRICS
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListMetricsQuery(final ListQuery query) {
		// TODO: Implement
		// FilterClause filterClause = ((ListQuery) query).getFilterClause();
		final ListQuery listQuery = query;
		final ListMetricsQuery listMetricsQuery = (ListMetricsQuery) query.getQueryType();

		final UsingClause usingClause = listQuery.getUsingClause();
		final ModelReferenceClause modelRefClause = usingClause.getModelReference();

		final EntityMapping mapping = mappingFactory.createEntityMapping();
		final EList<Resource> resources = mapping.getResources();
		final EList<Service> services = mapping.getServices();

		if (!getQueryConnector().supportsProbeRequests()) {
			log.error("Connector does not support probe requests, aborting.");

			return false;
		}

		mapping.setModelLocation(modelRefClause.getModel().getLocation());

		final EntityReferenceClause entityReferenceClause = listMetricsQuery.getEntityReferenceClause();

		if (entityReferenceClause == null) {
			log.error("Missing entitiy reference clause. Aborting.");
			return false;
		}

		final EList<EntityReference> refClause = entityReferenceClause.getEntityReferences();
		for (final EntityReference ref : refClause) {
			Entity entity = null;

			switch (ref.getType().getValue()) {
			case EntityType.RESOURCE_VALUE:
				entity = domainFactory.createResource();
				resources.add((Resource) entity);
				break;
			case EntityType.SERVICE_VALUE:
				entity = domainFactory.createService();
				services.add((Service) entity);
				break;
			default:
				log.error("Unknown model entity in request, aborting.");
				return false;
			}

			entity.setIdentifier(ref.getIdentifier());

			if (ref.getAlias() != null) {
				entity.setAlias(ref.getAlias().getAlias());
			}
		}

		setResult(getQueryConnector().requestProbes(mapping));

		return true;
	}

	/**
	 * Entry point for interpreting a LIST query
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListQuery(final ListQuery query) {
		final EObject queryType = query.getQueryType();

		// Decide which kind of LIST query is incoming and delegate its
		// processing to a suitable method
		if (queryType instanceof ListEntitiesQuery) {
			return interpretListEntitiesQuery(query);
		} else if (queryType instanceof ListResourcesQuery) {
			return interpretListResourcesQuery(query);
		} else if (queryType instanceof ListServicesQuery) {
			return interpretListServicesQuery(query);
		} else if (queryType instanceof ListMetricsQuery) {
			return interpretListMetricsQuery(query);
		} else if (queryType instanceof ListDoFQuery) {
			return interpretListDoFQuery(query);
		}

		log.error("No suitable interpreter found in interpretListQuery() for " + queryType);
		return false;
	}

	/**
	 * Interpret query as LIST RESOURCES
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListResourcesQuery(final ListQuery query) {
		// TODO: Ensure a filter INSTANCEOF/Resources is applied, Implement :-)
		final boolean state = interpretListEntitiesQuery(query);

		log.warn("LIST RESOURCES WORKAROUND: Clearing all lists containing services as workaround. Implement Filtering!");

		for (EntityMapping result : getResults()) {
			result.getServices().clear();
		}

		return state;
	}

	/**
	 * Interpret query as LIST SERVICES
	 * 
	 * @param query
	 *            The ListQuery to be down casted
	 * @return True, if interpreting was successful
	 */
	private boolean interpretListServicesQuery(final ListQuery query) {
		// TODO: Ensure a filter INSTANCEOF/Services is applied, Implement :-)
		final boolean state = interpretListEntitiesQuery(query);

		log.warn("LIST SERVICES WORKAROUND: Clearing all lists containing resources as workaround. Implement Filtering!");

		for (EntityMapping result : getResults()) {
			result.getResources().clear();
		}

		return state;
	}

	public boolean interpretQuery() {
		final ModelStructureQuery query = getQuery();
		
		if (query instanceof ListQuery) {
			final ListQuery listQuery = (ListQuery) query;
			final UsingClause usingClause = listQuery.getUsingClause();
			final ModelReferenceClause modelReference = usingClause.getModelReference();
			final ModelFamily family = modelReference.getFamily();
			final String familyIdentifier = family.getFamily();
			
			setModelFamily(familyIdentifier);

			if (!initQueryConnector()) {
				log.error("Could not load Connector for family " + getModelFamily() + ", aborting.");
				return false;
			}

			return interpretListQuery(listQuery);
		}

		log.error("Query type not supported");
		return false;
	}

}
