package edu.kit.ipd.descartes.ql.core.engine.interpreter;

import static edu.kit.ipd.descartes.ql.core.engine.observations.ObservationInterpreterFactory.createInterpreter;
import static edu.kit.ipd.descartes.ql.core.engine.util.DoFHelper.processVariationClause;
import static edu.kit.ipd.descartes.ql.core.engine.util.QueryParserHelper.toFlatConfigurationPropertiesClause;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

import edu.kit.ipd.descartes.ql.connector.PerformanceMetricsQueryConnector;
import edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateController;
import edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateControllerImpl;
import edu.kit.ipd.descartes.ql.core.engine.aggregates.AggregateProcessorImpl;
import edu.kit.ipd.descartes.ql.core.engine.observations.ObservationInterpreter;
import edu.kit.ipd.descartes.ql.core.engine.observations.SampleInterpreter;
import edu.kit.ipd.descartes.ql.lang.descartesQL.AggregateMetricClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.AliasClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ConfigurationPropertiesClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ConstraintClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.DoFClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.DoFReference;
import edu.kit.ipd.descartes.ql.lang.descartesQL.EntityReference;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ExplorationStrategyClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.MetricClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.MetricClauseType;
import edu.kit.ipd.descartes.ql.lang.descartesQL.MetricReference;
import edu.kit.ipd.descartes.ql.lang.descartesQL.MetricReferenceClauses;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelFamily;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelLocation;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ModelReferenceClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ObserveClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.PerformanceMetricsQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.SelectQuery;
import edu.kit.ipd.descartes.ql.lang.descartesQL.UsingClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.VaryingClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.WithClause;
import edu.kit.ipd.descartes.ql.models.mapping.domain.DoF;
import edu.kit.ipd.descartes.ql.models.mapping.domain.DomainFactory;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Entity;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Probe;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Resource;
import edu.kit.ipd.descartes.ql.models.mapping.domain.Service;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.EntityMapping;
import edu.kit.ipd.descartes.ql.models.mapping.mapping.MappingFactory;

/**
 * Interpreter for SELECT queries
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public class PerformanceMetricsQueryInterpreter extends
		AbstractQueryInterpreter<PerformanceMetricsQuery, PerformanceMetricsQueryConnector> {

	private final MappingFactory mappingFactory = MappingFactory.eINSTANCE;
	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;

	public PerformanceMetricsQueryInterpreter(final PerformanceMetricsQuery query) {
		super(query);
	}

	@Override
	protected Class<PerformanceMetricsQueryConnector> getQueryConnectorClass() {
		return PerformanceMetricsQueryConnector.class;
	}

	public boolean interpretQuery() {
		SelectQuery selectQuery = null;

		if (getQuery() instanceof SelectQuery) {
			selectQuery = (SelectQuery) getQuery();
			
			final UsingClause usingClause = selectQuery.getUsingClause();
			final ModelReferenceClause modelReference = usingClause.getModelReference();
			final ModelFamily family = modelReference.getFamily();
			final String familyIdentifier = family.getFamily();
			
			setModelFamily(familyIdentifier);

			if (!initQueryConnector()) {
				log.error("Could not load Connector for family " + getModelFamily() + ", aborting.");
				return false;
			}

			// Reset before work, allowing independent caching mechanisms in
			// the DQL Connector
			getQueryConnector().reset();

			return interpretSelectQuery(selectQuery);
		}

		log.error("Query type not supported");
		return false;
	}

	/**
	 * Entry point for interpreting SELECT
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @return True, if interpreting the query was successful
	 */
	private boolean interpretSelectQuery(final SelectQuery selectQuery) {
		final PerformanceMetricsQueryConnector queryConnector = getQueryConnector();
		final Map<String, Entity> aliasMap = new HashMap<String, Entity>();
		final EntityMapping request = mappingFactory.createEntityMapping();
		final AggregateController aggregates = new AggregateControllerImpl();
		final UsingClause usingClause = selectQuery.getUsingClause();
		final ModelReferenceClause modelReference = usingClause.getModelReference();
		final ModelLocation model = modelReference.getModel();
		final String location = model.getLocation();

		if (!queryConnector.supportsMetricRequests()) {
			log.error("Query connector has no support for Performance Metrics, aborting.");
			return false;
		}

		// process using
		request.setModelLocation(location);

		// process for
		if (!processForClause(selectQuery, aliasMap, request)) {
			return false;
		}

		// process select
		if (!processSelectClause(selectQuery, aliasMap, aggregates)) {
			return false;
		}

		// process constrained query
		if (queryConnector.supportsConstraintRequests()) {
			processConstraintClause(selectQuery, queryConnector);
		}

		// process observe clause
		processObserveClause(selectQuery, queryConnector);

		final boolean hasAggregates = aggregates.hasAggregates();
		final DoFClause dofClause = selectQuery.getDofClause();

		if (dofClause != null) {
			// Request results with the interpretation of DoF

			if (!queryConnector.supportsDoFRequests()) {
				log.error("Query connector has no support for DoF Queries, aborting.");
				return false;
			}

			processDoFClause(queryConnector, selectQuery, request);

			if (hasAggregates) {
				aggregates.preProcess(request, aliasMap);
			}

			setResults(queryConnector.requestDoFMetrics(request));

			if (hasAggregates) {
				for (final EntityMapping response : getResults()) {
					aggregates.postProcess(response);
				}
			}
		} else {
			// Request results without the interpretation of DoF

			if (hasAggregates) {
				aggregates.preProcess(request, aliasMap);
			}

			setResult(queryConnector.requestMetrics(request));

			if (hasAggregates) {
				aggregates.postProcess(getResult());
			}
		}

		return true;
	}

	/**
	 * Process a constraint clause in best-effort manner
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param queryConnector
	 *            The query connector to execute the query
	 */
	private void processConstraintClause(final SelectQuery selectQuery,
			final PerformanceMetricsQueryConnector queryConnector) {
		final ConstraintClause constraintClause = selectQuery.getConstraintClause();

		if (constraintClause != null) {
			final String constraint = constraintClause.getType();
			queryConnector.setConstraint(constraint);
		} else {
			queryConnector.setConstraint(null);
		}
	}

	/**
	 * Process a EVALUATE DOF clause with Degrees-of-Freedom
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param request
	 *            The request to be sent to the query connector
	 */
	private void processDoFClause(final PerformanceMetricsQueryConnector queryConnector, final SelectQuery selectQuery,
			final EntityMapping request) {
		final DoFClause doFClause = selectQuery.getDofClause();
		final VaryingClause varyingClause = doFClause.getVaryingClause();
		final ExplorationStrategyClause explorationStrategyClause = doFClause.getExplorationStrategy();
		final ConstraintClause constraintClause = selectQuery.getConstraintClause();
		final WithClause withClause = doFClause.getWithClause();

		if (withClause != null) {
			request.setDoFModelLocation(withClause.getModelReference().getModel().getLocation());
		}

		if (explorationStrategyClause != null) {
			processExplorationStrategyClause(queryConnector, explorationStrategyClause);
		}

		if (constraintClause != null) {
			queryConnector.setDoFConstraint(constraintClause.getType());
		}

		if (varyingClause != null) {
			processVaryingClause(request, varyingClause);
		}

	}

	/**
	 * Process a CONSTRAINED AS clause in best-effort manner
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param explorationStrategyClause
	 */
	private void processExplorationStrategyClause(final PerformanceMetricsQueryConnector queryConnector,
			final ExplorationStrategyClause explorationStrategyClause) {
		final ConfigurationPropertiesClause configurationProperties = explorationStrategyClause.getParameterClause();
		final Properties explorationStrategyProperties = toFlatConfigurationPropertiesClause(configurationProperties);
		queryConnector.setDoFExplorationStrategy(explorationStrategyClause.getName(), explorationStrategyProperties);
	}

	/**
	 * Process a FOR clause
	 * 
	 * @param queryConnector
	 *            The query connector to execute the query
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param aliasMap
	 *            A index structure with mappings from aliases to entities
	 * @param request
	 *            The request to be sent to the query connector
	 * 
	 * @return True, if the clause was consistent and could be interpreted
	 */
	private boolean processForClause(final SelectQuery selectQuery, final Map<String, Entity> aliasMap,
			final EntityMapping request) {
		final List<Resource> resources = request.getResources();
		final List<Service> services = request.getServices();
		final List<EntityReference> entityRefs = selectQuery.getForClause().getEntityReferenceClause()
				.getEntityReferences();

		for (final EntityReference ref : entityRefs) {
			Entity entity = null;

			switch (ref.getType()) {
			case RESOURCE:
				entity = domainFactory.createResource();
				resources.add((Resource) entity);
				break;
			case SERVICE:
				entity = domainFactory.createService();
				services.add((Service) entity);
				break;
			default:
				log.error("Invalid entity type found, aborting.");
				return false;
			}

			final AliasClause aliasClause = ref.getAlias();

			if (aliasClause != null) {
				entity.setAlias(aliasClause.getAlias());
			}

			entity.setIdentifier(ref.getIdentifier());

			if (aliasMap.put(entity.getAlias(), entity) != null) {
				log.error("Duplicate aliases in FOR, aborting.");
				return false;
			}
		}

		return true;
	}

	/**
	 * TODO: Alpha stage, skip in code review
	 * 
	 * @param selectQuery
	 * @param queryConnector
	 */
	private void processObserveClause(final SelectQuery selectQuery,
			final PerformanceMetricsQueryConnector queryConnector) {
		final ObserveClause observeClause = selectQuery.getObserveClause();
		final boolean supportsObservationPeriods = queryConnector.supportsObservationPeriods();

		if (observeClause == null) {
			return;
		}

		if (!supportsObservationPeriods) {
			log.warn("Observation period set, but not supported by the query connector. Ignoring OBSERVE.");
			return;
		}

		final ObservationInterpreter oi = createInterpreter(observeClause.getObservation());
		final SampleInterpreter si = new SampleInterpreter(observeClause.getSampleClause());

		oi.configureConnector(queryConnector, si.getSampleSize());
	}

	/**
	 * Process a SELECT clause an resolve all metrics and aggregates to be
	 * calculated
	 * 
	 * @param selectQuery
	 *            The query to be interpreted
	 * @param aliasMap
	 *            A index structure with mappings from aliases to entities
	 * @param aggregates
	 *            A controller for the processing of aggregates
	 * 
	 * @return True, if the clause was consistent and could be interpreted
	 */
	private boolean processSelectClause(final SelectQuery selectQuery, final Map<String, Entity> aliasMap,
			final AggregateController aggregates) {
		final MetricReferenceClauses metricRefClauses = selectQuery.getMetricReferenceClauses();
		final EList<MetricClauseType> metricClauseTypes = metricRefClauses.getMetricClauseTypes();

		for (final MetricClauseType ref : metricClauseTypes) {
			final Probe probe = domainFactory.createProbe();
			Entity entity = null;

			if (ref instanceof MetricClause) {
				// Found a plain metric
				final MetricClause metricClause = (MetricClause) ref;

				final MetricReference metricRef = metricClause.getMetricReference();
				final String entityRef = metricRef.getEntityReference();
				final String metricRef2 = metricRef.getMetricReference();

				entity = aliasMap.get(entityRef);
				probe.setMetricName(metricRef2);
			} else if (ref instanceof AggregateMetricClause) {
				// Found an aggregate
				aggregates.registerAggregate(new AggregateProcessorImpl((AggregateMetricClause) ref));
				continue;
			} else {
				log.error("Unsupported clause detected: " + ref);
				return false;
			}

			// Naming metrics from undefined aliases makes little sense
			if (entity == null) {
				log.error("Missing entity alias for requested metric, aborting.");
				return false;
			} else {
				final EList<Probe> probes = entity.getProbes();
				probes.add(probe);
			}
		}

		return true;
	}

	/**
	 * Process a VARYING clause to specify parameters as DoF
	 * 
	 * @param request
	 *            The request to be sent to the query connector
	 * @param varyingClause
	 *            The clause of parameters used as DoF
	 */
	private void processVaryingClause(final EntityMapping request, final VaryingClause varyingClause) {
		for (final DoFReference doFReference : varyingClause.getDofReferences()) {
			final EObject variationClause = doFReference.getDoFVariationClause();
			Properties variationProperties = null;
			final DoF doF = domainFactory.createDoF();
			final EList<DoF> requestDoF = request.getDoF();
			final AliasClause aliasClause = doFReference.getDoFAliasClause();
			
			if(aliasClause != null){
				doF.setAlias(aliasClause.getAlias());				
			}
			
			doF.setIdentifier(doFReference.getDoFIdentifier());

			if (variationClause != null) {
				variationProperties = processVariationClause(variationClause, variationProperties);

				doF.setProperties(variationProperties);
			}

			requestDoF.add(doF);
		}
	}
}
