package edu.kit.ipd.descartes.ql.core.engine.observations;

import java.util.Date;

import org.apache.log4j.Logger;

import edu.kit.ipd.descartes.ql.connector.PerformanceMetricsQueryConnector;
import edu.kit.ipd.descartes.ql.connector.util.DQLLogger;
import edu.kit.ipd.descartes.ql.lang.descartesQL.AbsoluteTimeClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.BetweenTimeClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.ObserveBetweenClause;
import edu.kit.ipd.descartes.ql.lang.descartesQL.RelativeTimeClause;

public class ObserveBetweenInterpreter implements ObservationInterpreter {
	private final Logger log = DQLLogger.getLogger(this.getClass().getName());
	private ObserveBetweenClause observation = null;
	private ObservationSpecification observationSpecification = null;

	public ObserveBetweenInterpreter(final ObserveBetweenClause observation) {
		this.observation = observation;
		interpretDate();
	}

	private TimeCalculator calcLeftAbsoluteRightAbsolute(final BetweenTimeClause endingTime) {
		// Right value can be calculated independently
		TimeCalculator endingTimeCalc;
		endingTimeCalc = new AbsoluteTimeCalculator((AbsoluteTimeClause) endingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftAbsoluteRightRelative(final BetweenTimeClause endingTime,
			final TimeCalculator startingTimeCalc) throws Exception {
		// Right value can be calculated independently
		TimeCalculator endingTimeCalc;
		if (!startingTimeCalc.isValid()) {
			// Without a valid starting time, the ending time cannot be
			// calculated
			throw new Exception();
		}

		endingTimeCalc = new RelativeTimeCalculator();
		endingTimeCalc.setReferenceDate(startingTimeCalc.getDate());
		endingTimeCalc.parseDate((RelativeTimeClause) endingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftRelativeRightAbsolute(final BetweenTimeClause startingTime,
			final BetweenTimeClause endingTime, final TimeCalculator startingTimeCalc) throws Exception {
		TimeCalculator endingTimeCalc;
		endingTimeCalc = calcLeftAbsoluteRightAbsolute(endingTime);

		if (!endingTimeCalc.isValid()) {
			throw new Exception();
		}

		startingTimeCalc.setReferenceDate(endingTimeCalc.getDate());
		startingTimeCalc.setBackwards(true);
		startingTimeCalc.parseDate((RelativeTimeClause) startingTime);
		return endingTimeCalc;
	}

	private TimeCalculator calcLeftRelativeRightRelative(final BetweenTimeClause startingTime,
			final BetweenTimeClause endingTime, final TimeCalculator startingTimeCalc) throws Exception {
		TimeCalculator endingTimeCalc;
		// Right value relative -> current time as reference for
		// negative relative starting time
		final Date currentTime = new Date();
		startingTimeCalc.setReferenceDate(currentTime);
		startingTimeCalc.setBackwards(true);
		startingTimeCalc.parseDate((RelativeTimeClause) startingTime);

		endingTimeCalc = calcLeftAbsoluteRightRelative(endingTime, startingTimeCalc);
		return endingTimeCalc;
	}

	public void configureConnector(final PerformanceMetricsQueryConnector queryConnector, final long sampleSize) {
		if (observationSpecification == null) {
			log.error("No observation specification set.");
			return;
		}

	}

	public void interpretDate() {
		final BetweenTimeClause startingTime = observation.getStartingTime();
		final BetweenTimeClause endingTime = observation.getEndingTime();
		TimeCalculator startingTimeCalc = null;
		TimeCalculator endingTimeCalc = null;
		boolean leftAbsolute = false;
		boolean rightAbsolute = false;

		if (startingTime instanceof AbsoluteTimeClause) {
			leftAbsolute = true;
		}

		if (endingTime instanceof AbsoluteTimeClause) {
			rightAbsolute = true;
		}

		try {

			if (leftAbsolute) {
				// left is absolute
				startingTimeCalc = calcLeftAbsoluteRightAbsolute(startingTime);

				if (rightAbsolute) {
					endingTimeCalc = calcLeftAbsoluteRightAbsolute(endingTime);
				} else {
					endingTimeCalc = calcLeftAbsoluteRightRelative(endingTime, startingTimeCalc);
				}

			} else {
				// left is relative
				startingTimeCalc = new RelativeTimeCalculator();

				if (rightAbsolute) {
					endingTimeCalc = calcLeftRelativeRightAbsolute(startingTime, endingTime, startingTimeCalc);
				} else {
					endingTimeCalc = calcLeftRelativeRightRelative(startingTime, endingTime, startingTimeCalc);
				}

			}

			if (startingTimeCalc.isValid() && endingTimeCalc.isValid()) {
				storeResult(startingTimeCalc, endingTimeCalc);
			}

		} catch (final Exception e) {
			return;
		}
	}

	private void storeResult(final TimeCalculator startingTimeCalc, final TimeCalculator endingTimeCalc) {
		final Date startDate = startingTimeCalc.getDate();
		final Date endDate = endingTimeCalc.getDate();

		observationSpecification = new ObservationSpecification(startDate, endDate);
	}

}
