package edu.kit.ipd.descartes.ql.core.registry;

import java.util.Set;

import org.osgi.framework.ServiceReference;

import edu.kit.ipd.descartes.ql.connector.ConnectorProvider;

/**
 * Interface of of a central registry of query connectors for specific
 * performance meta-model families.
 * 
 * The implementation is based on OSGi Declarative Services as described by the
 * OSGi Standards. The usage scenario for this service is to provide a central
 * entry point in the OSGi run-time to search for query connectors for a
 * particular performance meta-model family. The query connectors are referenced
 * through OSGi Service References. Without this central data structure, a class
 * requiring a query connector would have to instantiate each query connector to
 * check for its support of a specific family.
 * 
 * @see META-INF/component.xml
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 */
public interface ConnectorRegistry {
	/**
	 * Method used by all classes that implement
	 * {@link edu.kit.ipd.descartes.ql.connector.ConnectorProvider} to register
	 * with the connector registry
	 * 
	 * @see edu.kit.ipd.descartes.ql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void bindConnector(ServiceReference<ConnectorProvider> serviceReference);

	/**
	 * Returns a set of names for registered connectors
	 * 
	 * @return The set of registered connector names
	 */
	public Set<String> getRegisteredConnectorNames();

	/**
	 * Method used by the query execution engine to create a reference to a
	 * query connector for a specific performance meta-model family
	 * 
	 * @see edu.kit.ipd.descartes.ql.core.engine.query.QueryExecutionEngine
	 * @param familyIdentifier
	 *            The identifier of the performance meta-model family
	 * @param queryConnector
	 *            The query class for which a query connector is required
	 * @return
	 */
	public <S, T> ServiceReference<S> requestServiceReference(String familyIdentifier, Class<T> queryConnector);

	/**
	 * Method used by all classes that implement
	 * {@link edu.kit.ipd.descartes.ql.connector.ConnectorProvider} to
	 * unregister with the connector registry
	 * 
	 * @see edu.kit.ipd.descartes.ql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void unbindConnector(ServiceReference<ConnectorProvider> serviceReference);

	/**
	 * Method used by all classes that implement
	 * {@link edu.kit.ipd.descartes.ql.connector.ConnectorProvider} to update
	 * the connector registry
	 * 
	 * @see edu.kit.ipd.descartes.ql.connector.ConnectorProvider
	 * @param serviceReference
	 */
	public void updateConnector(ServiceReference<ConnectorProvider> serviceReference);
}
