package edu.kit.ipd.descartes.ql.core.registry.impl;

import static edu.kit.ipd.descartes.ql.connector.ConnectorProvider.FAMILY_KEY;
import static edu.kit.ipd.descartes.ql.connector.ConnectorProvider.MODEL_STRUCTURE_SUPPORT_KEY;
import static edu.kit.ipd.descartes.ql.connector.ConnectorProvider.PERFORMANCE_ISSUES_SUPPORT_KEY;
import static edu.kit.ipd.descartes.ql.connector.ConnectorProvider.PERFORMANCE_METRICS_SUPPORT_KEY;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.osgi.framework.ServiceReference;

import edu.kit.ipd.descartes.ql.connector.ConnectorProvider;
import edu.kit.ipd.descartes.ql.connector.ModelStructureQueryConnector;
import edu.kit.ipd.descartes.ql.connector.PerformanceIssueQueryConnector;
import edu.kit.ipd.descartes.ql.connector.PerformanceMetricsQueryConnector;
import edu.kit.ipd.descartes.ql.core.registry.ConnectorRegistry;

/**
 * Implementation of a central registry of query connectors for specific
 * performance meta-model families
 * 
 * The implementation of the registry is not yet type-safe. Future versions
 * should be implemented with type-safe handling for registering classes,
 * depending on usage scenarios.
 * 
 * @author Fabian Gorsler <fabian.gorsler@student.kit.edu>
 * 
 */
public class ConnectorRegistryImpl implements ConnectorRegistry {
	@SuppressWarnings("rawtypes")
	private static Map<String, Map<Class, ServiceReference<ConnectorProvider>>> registry;

	@SuppressWarnings("rawtypes")
	protected synchronized static void initializeRegistry() {
		if (registry == null) {
			registry = new HashMap<String, Map<Class, ServiceReference<ConnectorProvider>>>();
		}
	}

	private Logger log = null;

	public ConnectorRegistryImpl() {
		log = DQLLogger.getLogger(this.getClass().getName());
		log.info("Registry starting");
	}

	@SuppressWarnings("rawtypes")
	public synchronized void bindConnector(final ServiceReference<ConnectorProvider> serviceReference) {
		log.info("Registering new connector");

		final String familyIdentifier = (String) serviceReference.getProperty(FAMILY_KEY);
		final String modelStructureSupportProperty = (String) serviceReference.getProperty(MODEL_STRUCTURE_SUPPORT_KEY);
		final String performanceMetricsSupportProperty = (String) serviceReference
				.getProperty(PERFORMANCE_METRICS_SUPPORT_KEY);
		final String performanceIssueSupportProperty = (String) serviceReference
				.getProperty(PERFORMANCE_ISSUES_SUPPORT_KEY);

		final boolean supportsModelStructure = Boolean.valueOf(modelStructureSupportProperty);
		final boolean supportsPerformanceMetrics = Boolean.valueOf(performanceMetricsSupportProperty);
		final boolean supportsPerformanceIssues = Boolean.valueOf(performanceIssueSupportProperty);

		Map<Class, ServiceReference<ConnectorProvider>> familyConnectors = registry.get(familyIdentifier);

		if (familyConnectors == null) {
			log.info("Creating new connector registry for " + familyIdentifier);
			familyConnectors = new HashMap<Class, ServiceReference<ConnectorProvider>>();
			registry.put(familyIdentifier, familyConnectors);
		}

		if (supportsModelStructure) {
			registerQueryClass(familyConnectors, serviceReference, ModelStructureQueryConnector.class);
		}

		if (supportsPerformanceMetrics) {
			registerQueryClass(familyConnectors, serviceReference, PerformanceMetricsQueryConnector.class);
		}

		if (supportsPerformanceIssues) {
			registerQueryClass(familyConnectors, serviceReference, PerformanceIssueQueryConnector.class);
		}
	}

	public Set<String> getRegisteredConnectorNames() {
		return registry.keySet();
	}

	@SuppressWarnings("rawtypes")
	private void registerQueryClass(final Map<Class, ServiceReference<ConnectorProvider>> familyConnectors,
			final ServiceReference<ConnectorProvider> serviceReference, final Class queryClass) {

		log.info("Registering " + queryClass + " for family "
				+ serviceReference.getProperty(ConnectorProvider.FAMILY_KEY));
		familyConnectors.put(queryClass, serviceReference);
	}

	@SuppressWarnings("unchecked")
	public <S, T> ServiceReference<S> requestServiceReference(final String familyIdentifier,
			final Class<T> queryConnector) {
		return (ServiceReference<S>) registry.get(familyIdentifier).get(queryConnector);
	}

	@SuppressWarnings("rawtypes")
	public synchronized void unbindConnector(final ServiceReference<ConnectorProvider> serviceReference) {
		log.info("Deregistering connector");

		final String familyIdentifier = (String) serviceReference.getProperty(ConnectorProvider.FAMILY_KEY);
		final Map<Class, ServiceReference<ConnectorProvider>> familyConnectors = registry.get(familyIdentifier);

		unregisterQueryClass(familyConnectors, serviceReference, ModelStructureQueryConnector.class,
				ConnectorProvider.MODEL_STRUCTURE_SUPPORT_KEY);
		unregisterQueryClass(familyConnectors, serviceReference, PerformanceMetricsQueryConnector.class,
				ConnectorProvider.PERFORMANCE_METRICS_SUPPORT_KEY);
		unregisterQueryClass(familyConnectors, serviceReference, PerformanceIssueQueryConnector.class,
				ConnectorProvider.PERFORMANCE_ISSUES_SUPPORT_KEY);

		if (registry.get(familyIdentifier).size() == 0) {
			log.info("Unregistering family " + familyIdentifier);
			registry.remove(familyIdentifier);
		}

	}

	private void unregisterQueryClass(
			@SuppressWarnings("rawtypes") final Map<Class, ServiceReference<ConnectorProvider>> familyConnectors,
			final ServiceReference<ConnectorProvider> serviceReference,
			@SuppressWarnings("rawtypes") final Class queryClass, final String queryClassKey) {
		if (!serviceReference.getProperty(queryClassKey).equals("true")) {
			// Only execute remove operation, if this query class has been
			// registered
			return;
		}

		log.info("Unregistering " + queryClass + " for family "
				+ serviceReference.getProperty(ConnectorProvider.FAMILY_KEY));
		familyConnectors.remove(queryClass);
	}

	public void updateConnector(final ServiceReference<ConnectorProvider> serviceReference) {
		// Up to now only bind and unbind are relevant events, but for
		// completeness this method is implemented as a stub
		log.info("Updating connector");
	}

}
